from uuid import uuid4
from urllib import parse

def uuid():
    return str(uuid4())

def merge(dict1, dict2):
    copy = dict1.copy()
    copy.update(dict2)
    return copy

def deep_obfuscate(payload, fields, obfuscate_name='[HIDDEN]'):
    try:
        for key in payload:
            if isinstance(key, str) and key.lower() in fields:
                payload[key] = obfuscate_name
            elif isinstance(payload, dict) and isinstance(payload[key], dict):
                payload[key] = deep_obfuscate(payload[key], fields, obfuscate_name)
            elif isinstance(payload, dict) and isinstance(payload[key], list):
                payload[key] = [deep_obfuscate(value, fields, obfuscate_name) for value in payload[key]]
        return payload
    except (AttributeError, TypeError):
        return payload

def deep_compact(payload):
    try:
        for key in payload.copy():
            if payload[key] is None or payload[key] == {} or payload[key] == []:
                del payload[key]
            elif isinstance(payload[key], dict):
                payload[key] = deep_compact(payload[key])
            elif isinstance(payload[key], list):
                payload[key] = [deep_compact(value) for value in payload[key]]
        return payload
    except (AttributeError, TypeError):
        return payload


def url_obfuscate(url, fields, obfuscate_name='HIDDEN'):
    uri = parse.urlsplit(url)
    query = deep_obfuscate(dict(parse.parse_qsl(uri.query)), fields, obfuscate_name)

    return parse.urljoin(url, '?{0}'.format(parse.urlencode(query)))



import re

from hawkei.store import Store
from hawkei.utils import url_obfuscate, uuid
from hawkei.config import config

_COOKIE_NAME = '_hawkei_stid'

class HawkeiMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if (config.active()):
            Store.init()
            Store.add({'session_tracker_id': self.get_session_tracker_id(request)})
            Store.add({'request': self.get_request_data(request)})
            Store.add({'software': request.META.get('SERVER_SOFTWARE')})

        response = self.get_response(request)

        if (config.active()):
            self.write_session_tracker_id(response)
            Store.delete()

        return response

    @classmethod
    def get_session_tracker_id(cls, request):
        return request.COOKIES.get(_COOKIE_NAME) or uuid()

    @classmethod
    def write_session_tracker_id(cls, response):
        response.set_cookie(
            _COOKIE_NAME,
            value=Store.get().get('session_tracker_id'),
            max_age=(10 * 365 * 24 * 60 * 60),
            path='/',
            httponly=False,
            domain=config.data.get('domain'),
        )

    @classmethod
    def get_request_data(cls, request):
        return ({
            'url': url_obfuscate(request.build_absolute_uri(), config.data['obfuscated_fields']),
            'get_params': dict(request.GET),
            'post_params': dict(request.POST),
            'ssl': request.scheme == 'https',
            'host': request.get_host(),
            'path': request.path,
            'referrer': url_obfuscate(request.META.get('HTTP_REFERER'), config.data['obfuscated_fields']),
            'method': request.method,
            'xhr': request.META.get('HTTP_X_REQUESTED_WITH') is not None,
            'user_agent': request.META.get('HTTP_USER_AGENT'),
            'ip': request.META.get('HTTP_X_FORWARDED_FOR') or request.META.get('REMOTE_ADDR'),
            'headers': cls.get_headers(request.META),
        })

    @classmethod
    def get_headers(cls, meta):
        pattern = re.compile("^HTTP_|^CONTENT_TYPE$|^CONTENT_LENGTH$")
        skip_headers=['HTTP_COOKIE']
        headers = {}

        for key in meta:
            if isinstance(key, str) and pattern.match(key) and not key in skip_headers:
                formated_key = '-'.join([value.title() for value in key.split('_')])
                headers[formated_key] = meta.get(key)

        return headers



from datetime import datetime, timezone
import socket
import os
import sys

from hawkei.version import VERSION
from hawkei.utils import merge, uuid, deep_compact
from hawkei.store import Store
from hawkei.config import config

_session_tracer_id = uuid()

def message_base():
    return ({
        'message_id': uuid(),
        'timestamp': datetime.utcnow().replace(tzinfo=timezone.utc).isoformat(),
        'session_tracker_id': _session_tracer_id,
    })

def message_extended():
    message = merge({
        'library': _library(),
        'server': _server(),
        'environment': config.data.get('environment_name'),
        'metadata': config.data.get('metadata'),
    }, message_base())

    message = merge(message, Store.get())

    return deep_compact(message)

def _library():
    return ({
        'name': 'hawkei',
        'language': 'python',
        'version': VERSION,
    })

def _server():
    return {
        'host': socket.gethostname(),
        'pid': os.getpid(),
        'software': Store.get().get('software') or sys.argv[0] or 'unknown',
    }



import logging
import atexit
import queue

from hawkei.processor.worker import Worker

_MAX_QUEUE_SIZE = 10000

class Async():
    logger = logging.getLogger('hawkei')

    def __init__(self):
        self.queue = queue.Queue(_MAX_QUEUE_SIZE)
        self.worker = Worker(self.queue)

        atexit.register(self.shutdown)
        self.worker.start()

    def enqueue(self, message):
        try:
            self.queue.put(message, block=False)
            return True

        except queue.Full:
            return False

    def flush(self):
        self.logger.debug('flush')
        self.queue.join()

    def shutdown(self):
        self.logger.debug('start stopping worker ...')
        self.worker.stop()
        self.flush()
        try:
            self.worker.join()
        except RuntimeError:
            pass



import logging

from hawkei.message import message_base, message_extended

from hawkei.utils import merge
from hawkei.config import config
from hawkei.processor.asynchronous import Async

_TRACK_EVENTS_CREATE = 'track_events:create'
_IDENTIFY_EVENTS_CREATE = 'identify_events:create'
_GROUP_EVENTS_CREATE = 'group_events:create'
_IDENTITIES_UPDATE = 'identities:update'
_WATCHERS_CREATE = 'watchers:create'

_processor = None

def init(user_config=None):
    config.build(user_config)

    global _processor # pylint: disable=global-statement
    _processor = Async()

def track(name, payload=None, options=None):
    if not config.active(): return False

    payload = payload or {}
    options = options or {}

    payload['name'] = name

    _processor.enqueue({
        'action': _TRACK_EVENTS_CREATE,
        'payload': merge(message_extended(), payload),
        'options': options,
    })
    return True

def identify(user_id, payload=None, options=None):
    if not config.active(): return False

    payload = payload or {}
    options = options or {}

    payload['user_id'] = user_id

    _processor.enqueue({
        'action': _IDENTIFY_EVENTS_CREATE,
        'payload': merge(message_extended(), payload),
        'options': options,
    })
    return True

def group(group_id, payload=None, options=None):
    if not config.active(): return False

    payload = payload or {}
    options = options or {}

    payload['group_id'] = group_id

    _processor.enqueue({
        'action': _GROUP_EVENTS_CREATE,
        'payload': merge(message_base(), payload),
        'options': options,
    })
    return True

def watch(flow, payload=None, options=None):
    if not config.active(): return False

    payload = payload or {}
    options = options or {}

    payload['template_flow'] = flow

    _processor.enqueue({
        'action': _WATCHERS_CREATE,
        'payload': merge(message_base(), payload),
        'options': options,
    })
    return True

def flush():
    if not config.active(): return False
    _processor.flush()
    return True

def shutdown():
    if not config.active(): return False
    _processor.shutdown()
    return True



from hawkei.store import Store

def test_init_store():
    Store.init()

    assert Store.get() == {}

def test_store_add():
    Store.init()

    assert Store.get() == {}
    assert Store.add({'test': True}) == {'test': True}
    assert Store.get() == {'test': True}

def test_store_clear():
    Store.init()

    assert Store.add({'test': True}) == {'test': True}
    assert Store.clear() == {}
    assert Store.get() == {}

def test_store_delete():
    Store.init()
    Store.delete()

    try:
        Store._store.val
    except AttributeError:
        assert True
    else:
        assert False

def test_get_without_init_store():
    assert Store.get() == {}

def test_clear_without_init_store():
    assert Store.clear() == {}

def test_add_without_init_store():
    assert Store.add({'test': True}) == {}



import logging
from threading import Thread
from queue import Empty
import backoff

from hawkei.processor.batch import Batch
from hawkei.request import create, ApiError
from hawkei.resources.batch import Batch as BatchResource

_INTERVAL = 3
_MAX_RETRIES = 10

class Worker(Thread):
    logger = logging.getLogger('hawkei')

    def __init__(self, queue):
        Thread.__init__(self)

        self.queue = queue
        self.daemon = True
        self.running = True
        self.batch = Batch()

    def run(self):
        self.logger.debug('worker is running ...')

        while self.running:
            try:
                message = self.queue.get(block=True, timeout=_INTERVAL)
                self.batch.add(message)
                if self.batch.is_full():
                    self.send_batch()

            except Empty:
                if not self.batch.is_empty(): self.send_batch()
                continue

        self.logger.debug('worker stopped')

    def stop(self):
        self.running = False

    def send_batch(self):

        def check_exception(error):
            if isinstance(error, ApiError):
                return error.status_code in [401, 404, 422]
            return False

        @backoff.on_exception(backoff.expo, Exception, max_tries=_MAX_RETRIES, giveup=check_exception)
        def send():
            create(BatchResource, {'data': self.batch.messages})

        self.logger.debug('send batch with: %d messages', len(self.batch.messages))
        try:
            send()
        except Exception as ex: # pylint: disable=W0703
            self.logger.error(ex)
        finally:
            self.queue.task_done()
            self.batch = Batch()



import threading

from hawkei.utils import merge

class Store():
    _store = None

    @classmethod
    def init(cls):
        cls._store = threading.local()
        cls._store.val = {}

    @classmethod
    def get(cls):
        try:
            return cls._store.val
        except AttributeError:
            return {}

    @classmethod
    def add(cls, attributes):
        try:
            cls._store.val = merge(cls._store.val, attributes)
            return cls._store.val
        except AttributeError:
            return {}

    @classmethod
    def clear(cls):
        try:
            cls._store.val = {}
            return cls._store.val
        except AttributeError:
            return {}

    @classmethod
    def delete(cls):
        cls.clear()
        del cls._store



import os
import sys
from setuptools import setup, find_namespace_packages

sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'hawkei'))
from version import VERSION

with open('README.md', 'r') as fh:
    long_description = fh.read()

install_requires = [
    'requests>=2',
    'backoff>=1.8',
]

tests_require = [
    'pytest>=4',
]

setup(
    name='hawkei',
    version=VERSION,
    url='https://github.com/hawkei-io/hawkei-python',
    author='Hawkei',
    author_email='support@hawkei.io',
    maintainer='Hawkei',
    maintainer_email='support@hawkei.io',
    packages=find_namespace_packages(include=['hawkei', 'hawkei.*']),
    license='MIT License',
    install_requires=install_requires,
    tests_require=tests_require,
    description='Hawkei python integration',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from datetime import datetime, timezone
import json
from requests import sessions

from hawkei.config import config
from hawkei.utils import deep_obfuscate

_session = sessions.Session()

def create(resource, payload):
    payload['sent_at'] = datetime.utcnow().replace(tzinfo=timezone.utc).isoformat()
    payload = deep_obfuscate(payload, config.data['obfuscated_fields'])

    data = json.dumps(payload).encode()
    url = '{0}/{1}'.format(config.data['api_host'], _path(resource))

    request = _session.post(url, data=data, headers=_headers(), timeout=5)

    if request.status_code >= 200 and request.status_code < 300: return request
    if request.status_code == 400: raise ApiError(request.status_code, 'Bad response from server')
    if request.status_code == 401: raise ApiError(request.status_code, 'Unauthorized, please check your api key')
    if request.status_code == 404: raise ApiError(request.status_code, 'Requested Resource not found')
    if request.status_code == 408: raise ApiError(request.status_code, 'Server timeout, verify status of the server')
    if request.status_code == 422: raise ApiError(request.status_code, 'Invalid parameters')
    if request.status_code in [449, 444, 501, 502]: raise ApiError(request.status_code, 'Server Error, please check with Hawkei')

    raise ApiError(request.status_code, request.reason)

def _path(resource):
    return 'api/{0}/{1}'.format(config.data['api_version'], resource.name())

def _headers():
    return ({
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        'X-Api-Key': config.data['api_key'],
        'X-Space-Name': config.data['space_name'],
        'X-Environment-Name': config.data['environment_name'],
    })


class ApiError(Exception):

    def __init__(self, status_code, message):
        super(ApiError, self).__init__(message)
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return "{0}: {1}".format(self.status_code, self.message)



from hawkei.config import config
from hawkei.store import Store
from hawkei.message import message_base, message_extended

def test_message_base():
    message = message_base()

    assert message['message_id']
    assert message['timestamp']
    assert message['session_tracker_id']

def test_message_extended():
    config.build({
        'api_key': 'xx42',
        'space_name': 'Example',
        'environment_name': 'test',
    })

    message = message_extended()

    assert message['message_id']
    assert message['timestamp']
    assert message['session_tracker_id']
    assert message['environment'] == 'test'

    assert message['library']['name'] == 'hawkei'
    assert message['library']['language'] == 'python'
    assert message['library']['version']

    assert message['server']['host']
    assert message['server']['pid']
    assert message['server']['software']

    assert not message.get('request')

def test_message_extended_with_metadata():
    config.build({
        'api_key': 'xx42',
        'space_name': 'Example',
        'environment_name': 'test',
        'metadata': {
            'version': '1.0'
        }
    })

    message = message_extended()

    assert message['metadata'] == {'version': '1.0'}

def test_message_extended_with_store():
    Store.init()
    Store.add({
        'session_tracker_id': 't-1234',
        'request': {'id': '1234'},
    })

    config.build({
        'api_key': 'xx42',
        'space_name': 'Example',
        'environment_name': 'test',
    })

    message = message_extended()

    assert message['session_tracker_id'] == 't-1234'
    assert message['request'] == {'id': '1234'}



from hawkei.utils import merge, deep_obfuscate, deep_compact, url_obfuscate

def test_merge():
    dict1 = {'x': 1, 'y': 2}
    dict2 = {'x': 1, 'z': 3}

    result = merge(dict1, dict2)

    assert dict1 == {'x': 1, 'y': 2}
    assert dict2 == {'x': 1, 'z': 3}
    assert result == {'x': 1, 'y': 2, 'z': 3}


def test_deep_obfuscate():
    payload = {
        'password': 'test',
        'another_key': {
            'credit_card': '41111111',
            'valid_key': 'test',
        },
        'withArray': [{
            'credit_card': '41111111',
            'CREDIT_CARD': '41111111',
            'CReDIT_CaRD': '41111111',
            'valid_key': 'test',
        }],
    }

    expected = {
        'password': '[HIDDEN]',
        'another_key': {
            'credit_card': '[HIDDEN]',
            'valid_key': 'test',
        },
        'withArray': [{
            'credit_card': '[HIDDEN]',
            'CREDIT_CARD': '[HIDDEN]',
            'CReDIT_CaRD': '[HIDDEN]',
            'valid_key': 'test',
        }],
    }

    assert deep_obfuscate(payload, ['password', 'credit_card']) == expected

def test_deep_compact():
    payload = {
        'user_id': 'test',
        'another_key': None,
        'nested': {
            'user_id': 'test',
            'another_key': None,
            'host': False,
        },
        'with_array': [
            {},
            {'test': True, 'another_key': None},
        ],
    }

    expected = {
        'user_id': 'test',
        'nested': {
            'user_id': 'test',
            'host': False,
        },
        'with_array': [
            {},
            {'test': True},
        ],
    }

    assert deep_compact(payload) == expected

def test_url_obfuscate():
    url = 'http://test.com?password=1234&test=true'
    expected = 'http://test.com?password=HIDDEN&test=true'
    assert url_obfuscate(url, ['password', 'credit_card']) == expected

    url = 'http://test.com'
    expected = 'http://test.com'
    assert url_obfuscate(url, ['password', 'credit_card']) == expected

    url = 'http://test.com/test?test=true'
    expected = 'http://test.com/test?test=true'
    assert url_obfuscate(url, ['password', 'credit_card']) == expected



from hawkei.utils import merge

_DEFAULT_OBFUSCATED_FIELDS = [
    'password',
    'password_confirmation',
    'secret',
    'secret_token',
    'authenticity_token',
    'token',
    'api_key',
    'access_token',
    'credit_card_number',
    'cvv',
    'ccv',
    'csrfmiddlewaretoken',
]

_DEFAULT = {
    'api_key': None,
    'space_name': None,
    'environment_name': None,
    'api_host': 'https://api.hawkei.io',
    'api_version': 'v1',
    'enabled': True,
    'obfuscated_fields': [],
    'metadata': {},
    'domain': None,
}

class Config():

    def __init__(self):
        self.data = None

    def build(self, user_config):
        self.data = merge(_DEFAULT, user_config)
        self.data['obfuscated_fields'] += _DEFAULT_OBFUSCATED_FIELDS
        self.validate()

    def active(self):
        if not self.data:
            return False

        return self.data['enabled']

    def validate(self):
        self.require('api_key', str)
        self.require('space_name', str)
        self.require('environment_name', str)
        self.require('api_host', str)

    def require(self, field, field_type):
        value = self.data[field]

        if not isinstance(value, field_type):
            message = 'Invalid configuration for field: {0}'.format(field)
            raise AssertionError(message)

config = Config()



import logging
import json

MAX_BYTES = 204800 # 200Kb
MAX_MESSAGES = 100
MAX_MESSAGE_BYTES = 32768 # 32Kb

class Batch():
    logger = logging.getLogger('hawkei')

    def __init__(self):
        self.total_bytes = 0
        self.messages = []

    def add(self, message):
        message_size = len(json.dumps(message).encode())

        if self.is_max_message_reached(message_size):
            self.logger.error('message is too large to be sent')
            return False

        self.total_bytes += message_size
        self.messages.append(message)
        return True


    def is_empty(self):
        return len(self.messages) == 0

    def is_full(self):
        return self.is_max_messages_reached() or self.is_max_size_reached()

    def is_max_messages_reached(self):
        return len(self.messages) >= MAX_MESSAGES

    def is_max_size_reached(self):
        return self.total_bytes >= MAX_BYTES

    @classmethod
    def is_max_message_reached(cls, message_json_size):
        return message_json_size > MAX_MESSAGE_BYTES



from hawkei.config import Config

def test_base_config():
    config = Config()

    config.build({
        'api_key': 'xx42',
        'space_name': 'Example',
        'environment_name': 'test',
    })

    assert config.data['api_key'] == 'xx42'
    assert config.data['space_name'] == 'Example'
    assert config.data['environment_name'] == 'test'
    assert config.data['api_host'] == 'https://api.hawkei.io'
    assert config.data['api_version'] == 'v1'
    assert config.data['enabled']
    assert config.data['obfuscated_fields']
    assert config.data['metadata'] == {}
    assert config.data['domain'] is None

def test_override_config():
    config = Config()

    config.build({
        'api_key': 'xx42',
        'space_name': 'Example',
        'environment_name': 'test',
        'enabled': False,
        'api_host': 'http:/test.com'
    })

    assert config.data['api_key'] == 'xx42'
    assert config.data['space_name'] == 'Example'
    assert config.data['environment_name'] == 'test'
    assert config.data['api_host'] == 'http:/test.com'
    assert config.data['api_version'] == 'v1'
    assert not config.data['enabled']
    assert config.data['obfuscated_fields']
    assert config.data['metadata'] == {}
    assert config.data['domain'] is None

def test_invalid_config():
    config = Config()

    try:
        config.build({})
    except AssertionError:
        assert True
    else:
        assert False

def test_active():
    config = Config()

    assert not config.active()

    config.build({
        'api_key': 'xx42',
        'space_name': 'Example',
        'environment_name': 'test',
    })

    assert config.active()

    config.build({
        'api_key': 'xx42',
        'space_name': 'Example',
        'environment_name': 'test',
        'enabled': False,
    })

    assert not config.active()
